from .sql import SQL
from . import flask
from .input import input_float, input_int, input_string
from .locales import usd, locale
from .printer import printer
import logging

logging.getLogger("Zaraciccio").disabled = True
